/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef HAVE_CONFIG_H
#include <config.h>
#endif

#include <artnet.h> 
#include <artnet/packets.h> 

/*
#ifdef HAVE_LIBPTHREAD
#include <pthread.h>
#endif
*/

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>


#ifndef _defs_defined
#define _defs_defined

// first artnet specific stuff

// port to listen and send from
#define ARTNET_PORT 6454

#define ARTNET_STRING "Art-Net\0"

#define ARTNET_STRING_SIZE 8

#define ARTNET_VERSION 14
#define OEM_CODE 0xff

#define TTM_BEHAVIOUR_MASK 0x02
#define TTM_REPLY_MASK 0x01

#define PROGRAM_NO_CHANGE	0x7f
#define PROGRAM_DEFAULTS	0x00
#define PROGRAM_CHANGE_MASK	0x80

#define HIGH_NIBBLE 0xF0
#define LOW_NIBBLE 0x0F

#define STATUS_PROG_AUTH_MASK 0x30

#define PORT_STATUS_LPT_MODE	0x02
#define PORT_STATUS_SHORT	0x04
#define PORT_STATUS_ERROR	0x04
#define PORT_STATUS_DISABLED_MASK 0x08
#define PORT_STATUS_MERGE	0x08
#define PORT_STATUS_DMX_TEXT	0x10
#define PORT_STATUS_DMX_SIP	0x20
#define PORT_STATUS_DMX_TEST	0x40
#define PORT_STATUS_ACT_MASK 	0x80

#define PORT_DISABLE_MASK 0x01

#define MIN_PACKET_SIZE 10

#define MERGE_TIMEOUT_SECONDS 10

// non artnet specific
#define SA struct sockaddr

#define TRUE 1
#define FALSE 0

#define min(a,b)    ((a) < (b) ? (a) : (b))
#define max(a,b)    ((a) > (b) ? (a) : (b))



// these define the types of node that can exist
// note it's different from artnet_node_t
typedef enum {
	STNODE = 0x00,
	STSERVER = 0x01,
	STMEDIA = 0x02,
	STROUTE = 0x03,
	STBACKUP = 0x04,
	STCONFIG = 0x05
} node_style;

// the status of the node
typedef enum {
	ARTNET_OFF,
	ARTNET_STANDBY,
	ARTNET_ON
} node_status_t ;


// callbacks have a function and some user data
typedef struct {
	int (*fh)(artnet_node n, void *p, void * data) ;
	void *data;	
} callback_t ;

typedef struct {
	int (*fh)(artnet_node n, int portid, void * data) ;
	void *data;	
} dmx_callback_t ;


// struct to store callbacks
typedef struct {
	callback_t recv;
	callback_t send;
	callback_t poll;
	callback_t reply;
	callback_t dmx;
	callback_t address;
	callback_t input;
	callback_t todrequest;
	callback_t toddata;
	callback_t ipprog;
	dmx_callback_t dmx_c ;
} node_callbacks_t ;

// the node report codes
typedef enum {
	ARTNET_RCDEBUG,
	ARTNET_RCPOWEROK,
	ARTNET_RCPOWERFAIL,
	ARTNET_RCSOCKETWR1,
	ARTNET_RCPARSEFAIL,
	ARTNET_RCUDPFAIL,
	ARTNET_RCSHNAMEOK,
	ARTNET_RCLONAMEOK,
	ARTNET_RCDMXERROR,
	ARTNET_RCDMXUDPFULL,
	ARTNET_RCDMXRXFULL,
	ARTNET_RCSWITCHERR,
	ARTNET_RCCONFIGERR,
	ARTNET_RCDMXSHORT,
	ARTNET_RCFIRMWAREFAIL,
	ARTNET_RCUSERFAIL
} node_report_code ;

// struct to hold the state of the node
typedef struct {
	artnet_node_type node_type;
	node_status_t mode;
	struct in_addr reply_addr;
	struct in_addr ip_addr;
	struct in_addr bcast_addr ;
	uint8_t hw_addr[6] ;
	int send_apr_on_change;
	char short_name[ARTNET_SHORT_NAME_LENGTH] ;
	char long_name[ARTNET_LONG_NAME_LENGTH] ;
	char report[ARTNET_REPORT_LENGTH] ;
	int ar_count ;
	uint8_t subnet ;
	uint8_t default_subnet ;
	uint8_t subnet_net_ctl ;

	node_report_code report_code;
} node_state_t ;


typedef struct {
	uint8_t addr ;
	uint8_t default_addr ;
	uint8_t net_ctl ;
	uint8_t status ;
} port_t ;

/**
 * struct to represent an input port
 * input ports need to keep track of sequence numbers
 * (this doesn't seem to be used though)
 *
 */
typedef struct {
	port_t port;
	uint8_t seq ;
} input_port_t ;


/**
 * merge modes for output ports
 *
 */
typedef enum {
	ARTNET_MERGE_HTP,
	ARTNET_MERGE_LTP
} merge_t ;

/**
 * struct to represent an output port
 * 
 * output ports can merge data from two sources in either HTP 
 * (highest takes precedence) or LPT (lowest takes precedence) mode
 * 
 * we need to store:
 *   o The data from each source
 *   o The ip of the source
 *   o The time the data was recv'ed
 *
 */
typedef struct {
	port_t port;
	int	length;								// the length of the data THAT HAS CHANGED since the last dmx packet
	uint8_t enabled ;						//  true if the port has had it's address set, the is internal only, it's not used by the ArtNet protocol, otherwise the node keeps picking up packets for the 0x00 port
	uint8_t	data[ARTNET_MAX_DMX] ;			// output data
	// for merging
	merge_t merge_mode;
	uint8_t dataA[ARTNET_MAX_DMX] ;
	uint8_t dataB[ARTNET_MAX_DMX] ;
	time_t timeA ;
	time_t timeB ;
	struct in_addr ipA;
	struct in_addr ipB;
} output_port_t ;

#define port_addr port.addr
#define port_default_addr port.default_addr
#define port_net_ctl port.net_ctl
#define port_status port.status

/**
 * the node list stores a pointer to the first,last and current 
 * entries.
 *
 */
typedef struct {
	artnet_node_entry first;
	artnet_node_entry current;
	artnet_node_entry last;
	int length;	
} node_list_t ;


/**
 * the main node structure
 *
 */
typedef struct {
	int sd[2] ;									// the two sockets
	node_state_t state;							// the state struct
	node_callbacks_t callbacks;					// the callbacks struct
	struct ports_s {
		uint8_t	types[ARTNET_MAX_PORTS] ;		// type of port
		input_port_t in[ARTNET_MAX_PORTS] ;		// input ports
		output_port_t out[ARTNET_MAX_PORTS] ;	// output ports
	} ports;
	artnet_reply_t ar_temp ;					// buffered artpoll reply packet
	node_list_t node_list ;						// node list
} artnet_node_t ;


typedef artnet_node_t *node ;

#endif

// function defs follow

#ifndef _artnet_defined
#define _artnet_defined

void _build_art_poll_reply(node n) ;
#endif

#ifndef _errors_defined
#define _errors_defined

void err_warn(const char * fmt , ... ) ;
void err_fatal(const char *fmt, ... ) ;
void *Malloc(int len);
void *Calloc(size_t nmemb, int len);
int init_mutexes(node n) ;
int lock_node(node n) ;
int unlock_node(node n) ;
int16_t bytes_to_short(uint8_t h, uint8_t l) ;
int32_t bytes_to_32(uint8_t bytes[4]) ;
uint8_t short_get_low_byte(int16_t i) ;
uint8_t short_get_high_byte(int16_t i) ;

#endif

#ifndef _network_defined
#define _network_defined

ssize_t net_recv(node n, artnet_packet  p, int block) ;
void net_send(node n, artnet_packet p) ;
int net_set_non_block(node n) ;
int net_init(node n, const char *ip) ;
int net_start(node n) ;
int net_close(node n) ;

#endif


#ifndef _reader_defined
#define _reader_defined

int handle(node n, artnet_packet p ) ;
int16_t get_type(artnet_packet p ) ;

#endif


#ifndef _writer_defined
#define _writer_defined

void send_poll_reply(node n, int reply) ;
void send_tod_data(node n) ;

#endif
